# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["CooldownStatus", "Bucket", "UserBucket", "GuildBucket", "GlobalBucket", "ChannelBucket", "CooldownManager"]

import abc
import enum
import inspect
import time
import typing as t

from lightbulb import errors

if t.TYPE_CHECKING:
    from lightbulb.context import base as ctx_base


class CooldownStatus(enum.Enum):
    INACTIVE = enum.auto()
    ACTIVE = enum.auto()
    EXPIRED = enum.auto()


class Bucket(abc.ABC):
    """
    Base class for representing a cooldown bucket. All buckets should inherit from this class.

    Args:
        length (:obj:`float`): Length of the cooldown timer.
        max_usages (:obj:`int`): Number of command usages before the cooldown is activated.
    """

    __slots__ = ("length", "usages", "commands_run", "activated", "start_time")

    def __init__(self, length: float, max_usages: int) -> None:
        self.length = length
        self.usages = max_usages
        self.commands_run: int = 0
        """Commands run for this bucket since it was created."""
        self.activated = False
        self.start_time: t.Optional[float] = None
        """The start time of the bucket cooldown. This is relative to :meth:`time.perf_counter`."""

    @classmethod
    @abc.abstractmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        """
        Extracts the hash from the context which links a command usage to a single cooldown bucket.

        Args:
            context (:obj:`~.context.base.Context`): The context the command was invoked under.

        Returns:
            Hashable: Hashable object linking the context to a cooldown bucket.
        """
        ...

    def acquire(self) -> CooldownStatus:
        """
        Get the current state of the cooldown and add a command usage if the cooldown is not
        currently active or has expired.

        Returns:
            :obj:`~CooldownStatus`: The status of the cooldown bucket.
        """
        if self.active:
            return CooldownStatus.ACTIVE
        elif self.activated and self.expired:
            return CooldownStatus.EXPIRED

        self.commands_run += 1
        if self.commands_run >= self.usages:
            self.activated = True
            self.start_time = time.perf_counter()
        return CooldownStatus.INACTIVE

    @property
    def active(self) -> bool:
        """
        Whether or not the cooldown represented by the bucket is currently active.
        """
        return self.activated and not self.expired

    @property
    def expired(self) -> bool:
        """
        Whether or not the cooldown represented by the bucket has expired.
        """
        if self.start_time is not None:
            return time.perf_counter() >= (self.start_time + self.length)
        return True


class GlobalBucket(Bucket):
    """
    Global cooldown bucket. All cooldowns will be applied globally if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return 0


class UserBucket(Bucket):
    """
    User cooldown bucket. All cooldowns will be applied per user if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.author.id


class ChannelBucket(Bucket):
    """
    Channel cooldown bucket. All cooldowns will be applied per channel if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.channel_id


class GuildBucket(Bucket):
    """
    Guild cooldown bucket. All cooldowns will be applied per guild if you use this bucket. The command will
    still be permitted to be run in DMs in which case the DM channel ID will be used as the cooldown identifier
    instead of a guild ID.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.guild_id if context.guild_id is not None else context.channel_id


class CooldownManager:
    """
    The cooldown manager for a command.
    """

    __slots__ = ("callback", "cooldowns")

    def __init__(
        self,
        callback: t.Callable[[ctx_base.Context], t.Union[Bucket, t.Coroutine[t.Any, t.Any, Bucket]]],
    ) -> None:
        self.callback = callback
        self.cooldowns: t.MutableMapping[t.Hashable, Bucket] = {}
        """Mapping of a hashable to a :obj:`~Bucket` representing the currently stored cooldowns."""

    async def _get_bucket(self, context: ctx_base.Context) -> Bucket:
        bucket = self.callback(context)
        if inspect.iscoroutine(bucket):
            assert not isinstance(bucket, Bucket)
            bucket = await bucket
        assert isinstance(bucket, Bucket)
        return bucket

    async def add_cooldown(self, context: ctx_base.Context) -> None:
        """
        Add a cooldown under the given context. If an expired bucket already exists then it
        will be overwritten.

        Args:
            context (:obj:`~.context.base.Context`): The context to add a cooldown under.

        Returns:
            ``None``
        """
        bucket = await self._get_bucket(context)

        cooldown_hash = bucket.extract_hash(context)
        cooldown_bucket = self.cooldowns.get(cooldown_hash)

        if cooldown_bucket is not None:
            cooldown_status = cooldown_bucket.acquire()
            if cooldown_status is CooldownStatus.ACTIVE:
                # Cooldown has been activated
                assert cooldown_bucket.start_time is not None
                raise errors.CommandIsOnCooldown(
                    "This command is on cooldown",
                    retry_after=(cooldown_bucket.start_time + cooldown_bucket.length) - time.perf_counter(),
                )
            elif cooldown_status is CooldownStatus.INACTIVE:
                # Cooldown has not yet been activated.
                return

        self.cooldowns[cooldown_hash] = bucket
        self.cooldowns[cooldown_hash].acquire()

    async def reset_cooldown(self, context: ctx_base.Context) -> None:
        """
        Reset the cooldown under the given context.

        Args:
            context (:obj:`~.context.base.Context`): The context to reset the cooldown under.

        Returns:
            ``None``
        """
        bucket = await self._get_bucket(context)
        del self.cooldowns[bucket.extract_hash(context)]



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["serialise_command", "manage_application_commands"]

import collections
import logging
import typing as t

import hikari

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb.commands import base

_LOGGER = logging.getLogger("lightbulb.internal")


class _GuildIDCollection:
    __slots__ = ("ids",)

    def __init__(self, ids: t.Optional[t.Union[t.Sequence[int], int]]):
        if ids is None:
            ids = []
        elif isinstance(ids, int):
            ids = [ids]
        self.ids: t.List[int] = list(ids)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, int):
            return NotImplemented
        return other in self.ids

    def __repr__(self) -> str:
        return f"Guilds({', '.join(str(i) for i in self.ids)})"


def _serialise_option(option: hikari.CommandOption) -> t.Dict[str, t.Any]:
    return {
        "type": option.type,
        "name": option.name,
        "description": option.description,
        "is_required": option.is_required,
        "choices": list(
            sorted(
                [{"n": c.name, "v": c.value} for c in option.choices] if option.choices is not None else [],
                key=lambda d: d["n"],  # type: ignore
            )
        ),
        "options": [_serialise_option(o) for o in option.options] if option.options is not None else [],
        "channel_types": option.channel_types if option.channel_types is not None else [],
    }


def _serialise_hikari_command(command: hikari.Command) -> t.Dict[str, t.Any]:
    options = sorted(command.options or [], key=lambda o: o.name)
    return {
        "name": command.name,
        "description": command.description,
        "options": [_serialise_option(o) for o in options],
        "guild_id": command.guild_id,
    }


def _serialise_lightbulb_command(command: base.ApplicationCommand) -> t.Dict[str, t.Any]:
    create_kwargs = command.as_create_kwargs()
    return {
        "name": create_kwargs["name"],
        "description": create_kwargs["description"],
        "options": [_serialise_option(o) for o in sorted(create_kwargs["options"], key=lambda o: o.name)],  # type: ignore
        "guild_id": _GuildIDCollection(command.guilds) if command.guilds else None,
    }


def serialise_command(command: t.Union[hikari.Command, base.ApplicationCommand]) -> t.Dict[str, t.Any]:
    if isinstance(command, hikari.Command):
        return _serialise_hikari_command(command)
    return _serialise_lightbulb_command(command)


def _compare_commands(cmd1: base.ApplicationCommand, cmd2: hikari.Command) -> bool:
    return serialise_command(cmd1) == serialise_command(cmd2)


async def manage_application_commands(app: app_.BotApp) -> None:
    assert app.application is not None

    grouped_commands: t.Dict[t.Union[int, None], t.Dict[str, base.ApplicationCommand]] = collections.defaultdict(dict)
    for s_command in app._slash_commands.values():
        guilds = s_command.guilds or app.default_enabled_guilds or None
        if guilds is not None:
            for guild in guilds:
                grouped_commands[guild][s_command.name] = s_command
        else:
            grouped_commands[None][s_command.name] = s_command

    global_commands: t.Sequence[hikari.Command] = await app.rest.fetch_application_commands(app.application)
    _LOGGER.info("Processing global commands")
    for command in global_commands:
        registered_command = app.get_slash_command(command.name)
        if registered_command is None or registered_command.guilds:
            if app._delete_unbound_commands:
                _LOGGER.debug("Deleting global command %r as no implementation could be found", command.name)
                await command.delete()
            continue

        if not _compare_commands(registered_command, command):
            _LOGGER.debug("Recreating global command %r as it appears to have changed", command.name)
            await registered_command._auto_create()
        else:
            _LOGGER.debug("Not recreating global command %r as it does not appear to have changed", command.name)
            registered_command.instances[None] = command
        grouped_commands[None].pop(registered_command.name, None)

    all_guild_ids: t.Set[int] = set()
    all_guild_ids.update(app.default_enabled_guilds)
    for app_command in app._slash_commands.values():
        all_guild_ids.update(app_command.guilds)

    for guild_id in all_guild_ids:
        _LOGGER.info("Processing commands for guild %r", str(guild_id))
        guild_commands: t.Sequence[hikari.Command] = await app.rest.fetch_application_commands(
            app.application, guild_id
        )
        for command in guild_commands:
            registered_command = app.get_slash_command(command.name)

            if registered_command is None or not registered_command.guilds:
                if app._delete_unbound_commands:
                    _LOGGER.debug(
                        "Deleting command %r from guild %r as no implementation could be found",
                        command.name,
                        str(guild_id),
                    )
                    await command.delete()
                continue

            if not _compare_commands(registered_command, command):
                _LOGGER.debug(
                    "Recreating guild command %r in guild %r as it appears to have changed",
                    command.name,
                    str(guild_id),
                )
                await registered_command.create(guild_id)
            else:
                _LOGGER.debug(
                    "Not recreating guild command %r in guild %r as it does not appear to have changed",
                    command.name,
                    str(guild_id),
                )
                registered_command.instances[guild_id] = command
            grouped_commands[guild_id].pop(registered_command.name, None)

    for g_id, commands in grouped_commands.items():
        for app_cmd in commands.values():
            _LOGGER.debug(
                "Creating command %r %s as it does not seem to exist yet",
                app_cmd.name,
                f"in guild {g_id}" if g_id else "globally",
            )

            await app_cmd.create(g_id)
    _LOGGER.info("Command processing completed")
